/*
 * File:      serial.cpp
 * Purpose:   Serial port class method definitions
 * Author:    George Tasker
 *
 * Created:   30-05-1997
 *
 * Code contributions: Andrew Goldstein
 *
 */

/*
// SYNOPSIS START

   32bit Windows serial port handling classes

   Steps for normal usage of the wxSerial Class

      1) Create an instance of a wxSerial port
      2) Open the port
      3) Set the CallBack function to be used (optional)
      4) Enable the port
      5) Do desired reads/writes
      6) Close port

// SYNOPSIS STOP
*/

/* static const char sccsid[] = "%W% %G%"; */


#ifdef __GNUG__
	#pragma implementation "serial.h"
#endif


// For compilers that support precompilation, includes "wx.h".
#include  "wx/wxprec.h"

#ifdef    __BORLANDC__
#pragma hdrstop
#endif  //__BORLANDC__

#ifndef WX_PRECOMP
#include  <wx/wx.h>
#endif //WX_PRECOMP


#include "serial.h"

wxSerial_CallBack defaultwxSerialCallBack;
// forward declaration
DWORD FAR PASCAL CommWatchProc( LPSTR lpData );


wxChar GetParityChar(wxParity parity)
{
   wxChar parityChar;
   switch (parity)
   {
      case wxPARITY_NONE:
         parityChar = wxT('N');
         break;
      case wxPARITY_ODD:
         parityChar = wxT('O');
         break;
      case wxPARITY_EVEN:
         parityChar = wxT('E');
         break;
      case wxPARITY_MARK:
         parityChar = wxT('M');
         break;
      case wxPARITY_SPACE:
         parityChar = wxT('S');
         break;
      default:
         parityChar = wxT('N');
         break;
   }
   return parityChar;
}  // GetParityChar()


////////////////////////////////////////////////////////////////////
//
//  wxSerial Constructor
//                     
////////////////////////////////////////////////////////////////////
wxSerial::wxSerial(wxChar *name, bool showErrorMsgs)
{
    m_toggleRTS               = FALSE;
    m_showErrorMsg            = showErrorMsgs;

    m_info.dcb                = NULL;                   
    m_info.id                 = INVALID_HANDLE_VALUE;
    m_info.cb                 = NULL;
    m_info.portName           = NULL;

    m_info.hPostWnd           = NULL;
    m_info.connected          = FALSE;
    m_info.threadId           = 0;
    m_info.osWrite.Offset     = 0;
    m_info.osWrite.OffsetHigh = 0;
    m_info.osWrite.hEvent     = 0;
    m_info.osRead.Offset      = 0;
    m_info.osRead.OffsetHigh  = 0;
    m_info.osRead.hEvent      = 0;
    m_info.hPostEvent         = NULL;
    m_info.hCommWatchThread   = NULL;
    m_info.softEnable         = 0;

    Create (name);
}  // wxSerial::wxSerial constructor


////////////////////////////////////////////////////////////////////
//
//  wxSerial Destructor
//
////////////////////////////////////////////////////////////////////
wxSerial::~wxSerial()
{
    // close port if it has not been explicitly closed
    if (IsOpen())
        Close();

    // Free memory used by the name of the port (if any)
    if (m_info.portName)
        delete [] m_info.portName;
}  // wxSerial::~wxSerial destructor


////////////////////////////////////////////////////////////////////
//
// PRIVATE METHOD
//
// Generic function used for setting the error status when something
// fails during serial communication.
//
// If 'm_showErrorMsg' is true, the text of the error message will 
// actually be displayed
//
////////////////////////////////////////////////////////////////////
void wxSerial::setError(int errorCode)
{
    m_lastErrorCode = errorCode;

    if (m_showErrorMsg)
    {
        wxString ErrorMsg = GetErrorText();
        wxMessageBox(ErrorMsg, wxT("Error Notification"), wxOK);
    }
}  // wxSerial::setError()



////////////////////////////////////////////////////////////////////
//
// Returns the text version of the last wxSerial error to occur
//
// NOTE: Copy the return wxChar* value for long term use.
// 
////////////////////////////////////////////////////////////////////
const wxChar *wxSerial::GetErrorText()
{
   static wxString msg;

   switch (m_lastErrorCode)
   {
      case 0:
         msg = wxEmptyString;
         break;
      case wxSERIAL_PORT_IS_ALREADY_OPEN:
         msg = wxT("Port is already open.\n\nPort must be closed before being re-created.");
         break;
      case wxSERIAL_PORT_IS_NOT_OPEN:
         msg = wxT("Port has not been opened yet.");
         break;
      case wxSERIAL_PORT_IS_NOT_NAMED:
         msg = wxT("Port has not been named");
         break;
      case wxSERIAL_COULD_NOT_OPEN_PORT:
         msg.Printf(wxT("Could not open serial port: '%s'"),m_info.portName);
         break;
      case wxSERIAL_CANNOT_CREATE_PORT_SETTINGS:
         msg = wxT("Cannot create Device Control Block");
         break;
      case wxSERIAL_CANNOT_CHANGE_PORT_SETTINGS:
         msg = wxT("Cannot create Device Control Block");
         break;
      case wxSERIAL_CANNOT_ALLOCATE_BUFFERS:
         msg = wxT("Cannot allocate serial buffers");
         break;
      case wxSERIAL_READ_ERROR:
         msg.Printf(wxT("Problem reading from serial port: "),m_info.portName);
         break;

      // Windows specific error messages
      case wxSERIAL_CANNOT_CREATE_READ_EVENT:
         msg = wxT("Cannot create a Read event");
         break;
      case wxSERIAL_CANNOT_CREATE_WRITE_EVENT:
         msg = wxT("Cannot create a Write event");
         break;
      case wxSERIAL_CANNOT_CREATE_POST_EVENT:
         msg = wxT("Cannot create a Post event");
         break;
      case wxSERIAL_CANNOT_INIT_THREAD:
         msg = wxT("Cannot initialize the Thread in Enable()");
         break;
      default:
      {
         msg.Printf(wxT("Unknown error code encountered: %d"),m_lastErrorCode);
         break;
      }
   }

   return msg.c_str();

}  // wxSerial::GetErrorText()


////////////////////////////////////////////////////////////////////
//
// Allows changing of the port settings (baud,parity,wordlenm
// stopbits) without closing and re-opening the port
//
////////////////////////////////////////////////////////////////////
bool wxSerial::ChangeSettings(wxBaud baudRate, wxParity parity, int wordLen, int stopBits)
{
   wxString dcbStr;
   dcbStr.Printf(wxT("%s: BAUD=%d PARITY=%c DATA=%d STOP=%d"),
                 m_info.portName, baudRate, GetParityChar(parity), wordLen, stopBits); 

   if (!BuildCommDCB(dcbStr, m_info.dcb))
   {
      setError(wxSERIAL_CANNOT_CREATE_PORT_SETTINGS);
      return FALSE;
   }

   if (SetCommState(m_info.id, m_info.dcb)==0)
   {
      setError(wxSERIAL_CANNOT_CHANGE_PORT_SETTINGS);
      return FALSE;
   }

   return TRUE;
}  // wxSerial::ChangeSettings()



////////////////////////////////////////////////////////////////////
//
// Returns the current CommSettings last programmed for this port
//
////////////////////////////////////////////////////////////////////
void wxSerial::GetSettings(CommSettings *cs)
{
   cs->baudRate   = m_info.dcb->BaudRate;
   cs->parity     = m_info.dcb->Parity;
   cs->wordLen    = m_info.dcb->ByteSize;
   cs->stopBits   = m_info.dcb->StopBits;
}  // wxSerial::GetSettings()



////////////////////////////////////////////////////////////////////
//
// This method is called by the constructor for wxSerial and should
// not need to be called directly unless the port is closed and this
// instance of wxSerial is to be re-assigned to a different port.
//
////////////////////////////////////////////////////////////////////
bool wxSerial::Create(wxChar *name)
{
   if (m_info.id != INVALID_HANDLE_VALUE)
   {
      setError(wxSERIAL_PORT_IS_ALREADY_OPEN);
      return FALSE;
   }

   if (m_info.portName)
      delete [] m_info.portName;

   m_info.portName = new wxChar[strlen(name)+1];
   strcpy(m_info.portName,name);

   return TRUE;

}  // wxSerial::Create()

   
////////////////////////////////////////////////////////////////////
// This function will flush both buffers of any data, and close the 
// serial port so it may be used by other processes.
//
// NOTE: The serial port object still exists after being closed.  
//       It may be opened for use again using the Open() function.
////////////////////////////////////////////////////////////////////
bool wxSerial::Close()
{
   if (!IsOpen())
   {
      setError(wxSERIAL_PORT_IS_NOT_OPEN);
      return 0;
   }   

   m_info.connected = FALSE;
   
   // disable event notification
   SetCommMask( m_info.id, 0 );

   // murder thread
   if (m_info.hCommWatchThread)
      CloseHandle(m_info.hCommWatchThread);
// m_info.threadId != 0;

   // wait for thread to die
// while (m_info.threadId != 0);

   EscapeCommFunction(m_info.id, CLRDTR);

   // purge any outstanding reads/writes and close device handle
   PurgeComm( m_info.id, PURGE_TXABORT | PURGE_RXABORT | 
                           PURGE_TXCLEAR | PURGE_RXCLEAR );

   CloseHandle( m_info.osRead.hEvent );
   CloseHandle( m_info.osWrite.hEvent );
	CloseHandle( m_info.hPostEvent );
   CloseHandle( m_info.id );

   // clean up event objects
   delete m_info.dcb;
   m_info.id = INVALID_HANDLE_VALUE;  

   return TRUE;
}  // wxSerial::Close()



////////////////////////////////////////////////////////////////////
//
// Returns a value indicating whether the serial port has been 
// successfully opened for use
//
////////////////////////////////////////////////////////////////////
bool wxSerial::IsOpen()
{
   return (m_info.id != INVALID_HANDLE_VALUE);
} // wxSerial::IsOpen()



////////////////////////////////////////////////////////////////////
//
// Handles all the initialization of the serial port parameters.
//
// NOTE: The Create() function must be called before calling this 
//       function
//
////////////////////////////////////////////////////////////////////
bool wxSerial::Open(wxBaud baudRate, wxParity parity, int wordLen, int stopBits, int inSize, int outSize, int RTSdelay)
{
   if (IsOpen())
   {
      // Nothing needs to be done, since its already open
      return TRUE;
   }
   
   if (!m_info.portName)
   {
      setError(wxSERIAL_PORT_IS_NOT_NAMED);
      return FALSE;
   }

   m_info.id = CreateFile(m_info.portName,
                        GENERIC_READ | GENERIC_WRITE,
                        0, /* exclusive access */
                        NULL, /* no security attrs */
                        OPEN_EXISTING,
                        NULL, /*FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED, */
                        NULL);

   if (!IsOpen())
   {
      setError(wxSERIAL_COULD_NOT_OPEN_PORT);
      return FALSE;
   }
     
   m_info.dcb = new DCB;

   FillMemory(m_info.dcb, sizeof(DCB), 0);
   m_info.dcb->DCBlength = sizeof(DCB);

   m_info.dcb->fBinary          = TRUE;
   m_info.dcb->fParity          = TRUE;
   m_info.dcb->fDtrControl      = DTR_CONTROL_HANDSHAKE;
   m_info.dcb->fOutxCtsFlow     = FALSE;
   m_info.dcb->fOutxDsrFlow     = FALSE;
   m_info.dcb->fDsrSensitivity  = FALSE;
   m_info.dcb->fOutX            = FALSE;
   m_info.dcb->fInX             = FALSE;
   m_info.dcb->fRtsControl      = RTS_CONTROL_TOGGLE;

   if (!ChangeSettings(baudRate,parity,wordLen,stopBits))
      return FALSE;
/*
   wxString dcbStr;
   dcbStr.Printf("%s: BAUD=%d PARITY=%c DATA=%d STOP=%d",m_info.portName,baudRate,GetParityChar(parity),wordLen,stopBits); 
   wxMessageBox(dcbStr,"DCB configuration");
*/
   if (SetupComm( m_info.id, inSize, outSize))
      m_info.connected = TRUE;
   else
   {
      setError(wxSERIAL_CANNOT_ALLOCATE_BUFFERS);

      m_info.connected = FALSE;
      CloseHandle(m_info.id);
      m_info.id = INVALID_HANDLE_VALUE;

      return FALSE;
   }

   // create I/O event used for overlapped reads / writes
   m_info.osRead.hEvent = CreateEvent( NULL,   // no security
                                     TRUE,   // explicit reset req
                                     FALSE,  // initial event reset
                                     NULL);  // no name

   if (m_info.osRead.hEvent == NULL)
   {
      setError(wxSERIAL_CANNOT_CREATE_READ_EVENT);

      m_info.connected = FALSE;
      CloseHandle(m_info.id);
      m_info.id = INVALID_HANDLE_VALUE;

      return FALSE;
   }

   m_info.osWrite.hEvent = CreateEvent( NULL,     // no security
                                      TRUE,     // explicit reset req
                                      FALSE,    // initial event reset
                                      NULL );   // no name

   if (NULL == m_info.osWrite.hEvent)
   {
      setError(wxSERIAL_CANNOT_CREATE_WRITE_EVENT);

      m_info.connected = FALSE;
      CloseHandle(m_info.id);
      m_info.id = INVALID_HANDLE_VALUE;

      return FALSE;
   }

   // create "posted notification" event
   m_info.hPostEvent = CreateEvent( NULL,    // no security
                                    TRUE,    // manual reset
                                    TRUE,    // initial event is set
                                    NULL );  // no name

   if (m_info.hPostEvent == NULL)
   {
      setError(wxSERIAL_CANNOT_CREATE_POST_EVENT);

      CloseHandle( m_info.osRead.hEvent );
      CloseHandle( m_info.osWrite.hEvent );
      m_info.connected = FALSE;
      CloseHandle(m_info.id);
      m_info.id = INVALID_HANDLE_VALUE;

      return FALSE;
   }

   m_delayForRTS = RTSdelay;

   return TRUE;

}  // wxSerial::Open()



////////////////////////////////////////////////////////////////////
//
// Aborts pending data transmissions in any combination of the 
// transmit/receive buffers for the serial port
//
// 'buffer' can be:
//       wxSERIAL_TX                (transmit buffer)
//       wxSERIAL_RX                (receive buffer)
//       wxSERIAL_TX & wxSERIAL_RX  (transmit and receive buffers)
//
////////////////////////////////////////////////////////////////////
bool wxSerial::Abort(int buffer)
{
   bool result=TRUE;

   if (buffer & wxSERIAL_TX) 
      result = result && PurgeComm(m_info.id, PURGE_TXABORT);
   if (buffer & wxSERIAL_RX)
      result = result && PurgeComm(m_info.id, PURGE_RXABORT);

   return result;
}  // wxSerial::Abort()



////////////////////////////////////////////////////////////////////
//
// Sends a break signal out the port
//
////////////////////////////////////////////////////////////////////
bool wxSerial::Break(bool setting)
{
   int result;
   if (setting)
      result = EscapeCommFunction(m_info.id, SETBREAK);
   else
      result = EscapeCommFunction(m_info.id, CLRBREAK);

   return (result != 0);
}  // wxSerial::Break()



////////////////////////////////////////////////////////////////////
//
// Clears the contents transmit/receive buffers (in any combination)
// for the serial port
//
// 'buffer' can be:
//       wxSERIAL_TX                (transmit buffer)
//       wxSERIAL_RX                (receive buffer)
//       wxSERIAL_TX & wxSERIAL_RX  (transmit and receive buffers)
//
////////////////////////////////////////////////////////////////////
bool wxSerial::Flush(int buffer)
{
   bool result=TRUE;

   if (buffer & wxSERIAL_TX) 
      result = result && PurgeComm(m_info.id, PURGE_TXABORT | PURGE_TXCLEAR);
   if (buffer & wxSERIAL_RX)
      result = result && PurgeComm(m_info.id, PURGE_RXABORT | PURGE_RXCLEAR);

   return result;
}  // wxSerial::Flush()



////////////////////////////////////////////////////////////////////
//
// This version of ::Read allows you to specify a maximum number of
// bytes to be read in.  The bytes read are stored in 'data'
// 
// Return value is the number of bytes read and stored in 'data'.
//
// NOTE: There is no NULL character appended at the end of 'data', 
//       as a data stream may contain a NULL as an acceptable
//       character
//
////////////////////////////////////////////////////////////////////
int wxSerial::Read(wxChar *data, int maxLen)
{
   COMSTAT  ComStat;
   bool     fReadStat;
   DWORD    dwErrorFlags, dwLength;

   if ((int) m_info.id == -1)
      return 0;

   ClearCommError( m_info.id, &dwErrorFlags, &ComStat );

// dwLength = min( (DWORD) maxLen, ComStat.cbInQue );

   dwLength = maxLen;

   if (dwLength > 0)
   {
      fReadStat = (ReadFile( m_info.id, data, dwLength, &dwLength, /*&m_info.osRead overlapped*/ NULL) != 0);

      if (!fReadStat)
      {
         if (GetLastError() == ERROR_IO_PENDING)
         {
            // wait for a second for this transmission to complete
            if (WaitForSingleObject( m_info.osRead.hEvent, m_timeoutVal ))
               dwLength = 0;
            else
            {
               GetOverlappedResult( m_info.id,
                                    &m_info.osRead,
                                    &dwLength, FALSE );

               m_info.osRead.Offset += dwLength;
            }
         }
         else
            // some other error occurred
            dwLength = 0;
      }
   }

   return ((int) dwLength);
}  // wxSerial::Read()


////////////////////////////////////////////////////////////////////
//
// Will read either 'maxLen' characters into 'data', or will read
// an unknown number of characters in to 'data' until the 'termChar'
// character is encountered in the data stream.
//
// 'includeTerm' indicates whether the terminator character should be
// returned as the last character of valid 'data'.  If this is TRUE,
// the returned count value from this function will include the 
// terminator character in its count, otherwise it will not include
// the terminator character in the count.
//
// NOTE: Return value is the number of characters read from the port
//       and stored in 'data' (see 'includeTerm' above for additional
//       details).  There is no NULL character appended at the end of
//       'data', as a data stream may contain a NULL as an acceptable
//       character
//
////////////////////////////////////////////////////////////////////
int wxSerial::Read(wxChar *data, int maxLen, int termChar, bool includeTerm)
{
   int   dwLength;
   wxChar  rxStr[2];

   bool done=FALSE,abort=FALSE;

   rxStr[0] = 0;
   rxStr[1] = 0;
   dwLength = 0;

   while (!done && !abort)
   {
      abort = !Read(rxStr,1);

      if (!abort)
      {
         data[dwLength++]=rxStr[0];
         done = (dwLength >= maxLen) || (rxStr[0] == (wxChar)termChar);
      }
   }

   return dwLength;
} // wxSerial::Read()



////////////////////////////////////////////////////////////////////
//
// Writes a single character 'data' out the serial port.  
// 
// This function is called by the second Write() function (below) 
// which is used to send multiple byte messages out the serial port.
// If this byte is to be the last byte of a full length message that
// is being sent out the port, the 'lastByte' parameter should be set
// to TRUE so that the RTS line can be toggled (if set to be toggled).
// 
// Return value will be the number of bytes (0 or 1) that were written
// successfully out to the serial port
//
////////////////////////////////////////////////////////////////////
int wxSerial::Write(wxChar data, bool lastByte, bool calledFromWrite2)
{
   bool     fWriteStat;
   DWORD    dwBytesWritten;

   if (m_toggleRTS && !calledFromWrite2)
      SetRTS(TRUE);

   if (!IsOpen())
   {
      setError(wxSERIAL_PORT_IS_NOT_OPEN);
      return 0;
   }

   fWriteStat = (WriteFile(m_info.id, (LPSTR) &data, 1,
                   &dwBytesWritten, /*&m_info.osWrite overlapped*/ NULL) != 0);

   if (lastByte && m_toggleRTS && !calledFromWrite2)
   {
      // Need this to allow enough time for the last character to actually be sent out the port
      // before toggling the RTS line
      wxUsleep(m_delayForRTS);
      SetRTS(FALSE);
   }

   if (!fWriteStat && (GetLastError() == ERROR_IO_PENDING))
   {
      // wait for a short amount of time for this transmission to complete
      int value;
      if (value = WaitForSingleObject( m_info.osWrite.hEvent, m_timeoutVal ))
         dwBytesWritten = 0;
      else
      {
         GetOverlappedResult( m_info.id,
                              &m_info.osWrite,
                              &dwBytesWritten, FALSE );
         m_info.osWrite.Offset += dwBytesWritten;
      }
   }

   return(dwBytesWritten); 
}  // wxSerial::Write()


////////////////////////////////////////////////////////////////////
//
// Used to send multiple byte messages out a serial port
//
// Sends the first "len" bytes of "data" to port "port".  This 
// method will also handle blocks of bytes which contain NULLs, and
// send them out the port
//
// Return value is the number of bytes successfully written to the port.
//
////////////////////////////////////////////////////////////////////
int wxSerial::Write(DWORD len, const wxChar *data)
{
   DWORD dwBytesWritten = 0;

   if (m_toggleRTS)
      SetRTS(TRUE);

   WriteFile(m_info.id, data, len, &dwBytesWritten, /*&m_info.osWrite overlapped*/ NULL);

   if (m_toggleRTS)
   {
      // Need this to allow enough time for the last character to actually be sent out the port
      // before toggling the RTS line
      wxUsleep(m_delayForRTS);
      SetRTS(FALSE);
   }

   return dwBytesWritten;
}  // wxSerial::Write()



////////////////////////////////////////////////////////////////////
//
// Sends an XOFF transmission to the port
//
////////////////////////////////////////////////////////////////////
bool wxSerial::Xoff()
{
   int result;
   result = EscapeCommFunction(m_info.id, SETXOFF);

   return (result != 0);
}  // wxSerial:Xoff()



////////////////////////////////////////////////////////////////////
//
// Sends an XON transmission to the port
//
////////////////////////////////////////////////////////////////////
bool wxSerial::Xon()
{
   int result;
   result = EscapeCommFunction(m_info.id, SETXON);

   return (result != 0);
}  // wxSerial:Xon()



////////////////////////////////////////////////////////////////////
//
// Returns the number of bytes queued in the incoming buffer
//
////////////////////////////////////////////////////////////////////
DWORD wxSerial::PendingIn()
{
   DWORD    dwErrorFlags;
   COMSTAT  ComStat;

   ClearCommError( m_info.id, &dwErrorFlags, &ComStat );

   return(ComStat.cbInQue); 
}  // wxSerial::PendingIn()



////////////////////////////////////////////////////////////////////
//
// Returns the number of bytes queued to be sent out the port
//
////////////////////////////////////////////////////////////////////
DWORD wxSerial::PendingOut()
{
   DWORD    dwErrorFlags;
   COMSTAT  ComStat;

   ClearCommError( m_info.id, &dwErrorFlags, &ComStat );

   return(ComStat.cbOutQue); 
}  // wxSerial::PendingOut()


////////////////////////////////////////////////////////////////////
//
// Sets the DTR line according to 'setting'
//
////////////////////////////////////////////////////////////////////
bool wxSerial::SetDTR(bool setting)
{
   int result;
   if (setting)
      result = EscapeCommFunction( m_info.id, SETDTR);
   else
      result = EscapeCommFunction( m_info.id, CLRDTR);

   return (result != 0);
}  // wxSerial:SetDTR()



////////////////////////////////////////////////////////////////////
//
// Sets the RTS line according to 'setting'
//
////////////////////////////////////////////////////////////////////
bool wxSerial::SetRTS(bool setting)
{
   int result;
   if (setting)
      result = (EscapeCommFunction(m_info.id, SETRTS));
   else
      result = (EscapeCommFunction(m_info.id, CLRRTS));

   return (result != 0);
}  // wxSerial:SetRTS()



////////////////////////////////////////////////////////////////////
//
// Sets the delay that will be incurred before auto-toggling the RTS 
// line if 'toggleRTS' is set to TRUE.
//
////////////////////////////////////////////////////////////////////
void wxSerial::SetRTSDelay(int delay)
{
   m_delayForRTS = delay;
}  // wxSerial::SetRTSDelay()



////////////////////////////////////////////////////////////////////
//
// Sets the timeout period for the comm port for both reads and 
// writes to the port.  A one(1) millisecond gap is provided for
// the read and write of each character, plus the value passed in
// 'timeout'
//
////////////////////////////////////////////////////////////////////
bool wxSerial::SetTimeouts(DWORD timeout)
{
   COMMTIMEOUTS cto;
   GetCommTimeouts( m_info.id, &cto );
   
   cto.ReadTotalTimeoutConstant     = 0;
   cto.ReadIntervalTimeout          = timeout;
   cto.ReadTotalTimeoutMultiplier   = timeout;
   cto.WriteTotalTimeoutConstant    = 0;
   cto.WriteTotalTimeoutMultiplier  = timeout;

   // If using the port is opened in overlap mode, the above settings
   // are ignored.  This member variable of wxSerial holds the value
   // that is passed in to WaitForSingleObject() for a timeout
   m_timeoutVal = timeout;

   return (SetCommTimeouts( m_info.id, &cto ) != 0);
}  // wxSerial::SetTimeouts()



bool wxSerial::SetTimeouts(wxCommTimeouts timeouts)
{
   COMMTIMEOUTS cto;
   GetCommTimeouts( m_info.id, &cto );
   
   cto.ReadTotalTimeoutConstant     = timeouts.readTimeoutConstant;
   cto.ReadIntervalTimeout          = timeouts.readIntervalTimeout;
   cto.ReadTotalTimeoutMultiplier   = timeouts.readTimeoutMultiplier;
   cto.WriteTotalTimeoutConstant    = timeouts.writeTimeoutConstant;
   cto.WriteTotalTimeoutMultiplier  = timeouts.writeTimeoutMultiplier;

   // If the port is opened in overlap mode, the above settings
   // are ignored.  This member variable of wxSerial holds the value
   // that is passed in to WaitForSingleObject() for a timeout
   m_timeoutVal = timeouts.overlappedTimeout;

   return (SetCommTimeouts( m_info.id, &cto ) != 0);
}  // wxSerial::SetTimeouts()



////////////////////////////////////////////////////////////////////
//
// This function provides a way to set the private member variable 
// 'toggleRTS' to indicate whether the RTS line should be 
// automatically raised just before transmitting a telegram, and
// then lowered immediately after sending a telegram. 
//
// Default setting for the port is for 'toggleRTS' to be FALSE
//
////////////////////////////////////////////////////////////////////
void wxSerial::ToggleRTS(bool setting)
{
   m_toggleRTS = setting;
}  // wxSerial:ToggleRTS()



   
   
/*
 * ============================================================
 *
 *  Code for multi-threaded interrupt driven communication
 *
 *  This code was part of the original code that Andrew
 *  donated.  I (gt) have little knowledge of how to handle
 *  interrupt driven communication as far as integration into
 *  wxWindows, but it may be a starting point for someone else
 *
 * ============================================================
 */

/*
 * Used to create an initial thread for the serial port for interrupt driven
 * communication
 */
bool wxSerial::Enable(HWND hwnd)
{
   DWORD    dwThreadID;

   if (IsOpen())
   {
      setError(wxSERIAL_PORT_IS_NOT_OPEN);
      return FALSE;
   } 

   m_info.softEnable = 1;
// m_info.cb = &defaultwxSerialCallBack;
   
   if (NULL == (m_info.hCommWatchThread =
                     CreateThread(  (LPSECURITY_ATTRIBUTES) NULL,
                                    0, 
                                    (LPTHREAD_START_ROUTINE) CommWatchProc,
                                    (LPVOID) &m_info,
                                    0, &dwThreadID ))) 
   {
      setError(wxSERIAL_CANNOT_INIT_THREAD);

      m_info.connected = FALSE;
      CloseHandle(m_info.id);
      m_info.id = INVALID_HANDLE_VALUE;

      return FALSE;
   }
   else
   {
      m_info.hPostWnd = hwnd;
      m_info.threadId=dwThreadID;

      // Adjust thread priority
      SetThreadPriority( m_info.hCommWatchThread, THREAD_PRIORITY_BELOW_NORMAL);
      ResumeThread( m_info.hCommWatchThread );

      // assert DTR
      EscapeCommFunction( m_info.id, SETDTR );
   }
   
   return TRUE;
}  // wxSerial::Enable()


/*
 * Used to resume the functioning of an already defined thread that has been disabled/suspended
 */
void wxSerial::Enable()
{
   m_info.softEnable = 1;
   ResumeThread(m_info.hCommWatchThread);
}  // wxSerial::Enable()


/*
 * Suspends an existing thread
 */
void wxSerial::Disable()
{
   SuspendThread(m_info.hCommWatchThread);
   m_info.softEnable = 0;
}  // wxSerial::Disable()



void wxSerial::SetCallBack(wxSerial_CallBack *scb)
{
   m_info.cb = scb;
}  // wxSerial::SetCallBack()



wxSerial_CallBack::DoCallBack(wxSerial *port)
{
   COMSTAT  ComStat;
   DWORD    dwErrorFlags;

   ClearCommError( port->m_info.id, &dwErrorFlags, &ComStat );

   int count   =  ComStat.cbInQue;

   while(ComStat.cbInQue)
   {
      if (count)
      { 
         wxChar *data = new wxChar[count+1];
         int nRead = port->Read(data, count);
         if (nRead<0) 
            port->setError(wxSERIAL_READ_ERROR);
         else
            data[count]=0;

         delete data;
      }
      ClearCommError( port->m_info.id, &dwErrorFlags, &ComStat );
   }              

   return 0;
}  // wxSerial_CallBack::DoCallBack



DWORD FAR PASCAL CommWatchProc( LPSTR lpData )
{
   DWORD       dwTransfer, dwEvtMask;
   ThreadInfo  *info = (ThreadInfo*) lpData;
   OVERLAPPED  os;

   memset( &os, 0, sizeof( OVERLAPPED ) );

   // create I/O event used for overlapped read

   os.hEvent = CreateEvent( NULL,   // no security
                            TRUE,   // explicit reset req
                            FALSE,  // initial event reset
                            NULL ); // no name
   if (os.hEvent == NULL)
   {
      wxASSERT_MSG(0, wxT("Error: Cannot create thread event")); 
      return ( FALSE );
   }

   if (!SetCommMask( info->id, EV_RXCHAR ))
      return ( FALSE );

   while ( info->connected )                          
   {
      dwEvtMask = 0;           
      if (!WaitCommEvent( info->id, &dwEvtMask, &os ))
      {
         if (ERROR_IO_PENDING == GetLastError())
         {
            GetOverlappedResult( info->id, &os, &dwTransfer, TRUE );
            os.Offset += dwTransfer;
         }
      }

      if ((dwEvtMask & EV_RXCHAR) == EV_RXCHAR)
      {
         WaitForSingleObject( info->hPostEvent, 0xFFFFFFFF );

         // reset event
         ResetEvent( info->hPostEvent );

         if (info->softEnable)
         {
            // last message was processed, O.K. to post
            PostMessage( info->hPostWnd, WM_COMMNOTIFY,
                         (WPARAM) info->portName, MAKELONG( CN_EVENT, 0 ) );
         }
      }
   }

   info->threadId = 0   ;

   // get rid of event handle
   CloseHandle( os.hEvent );

   return( TRUE );

}  // CommWatchProc()
